"""Reachy parts submodule.

Used to define all parts in Reachy:

* :py:class:`RightArm`, :py:class:`LeftArm`
* :py:class:`~reachy.parts.hand.ForceGripper`
* :py:class:`Head`
"""

from .arm import LeftArm, RightArm # noqa
from .head import Head # noqa